#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_circleseq_std
----------------------------------

Tests for `circleseq` module.
"""

import yaml
import unittest
import os
import shutil
import utils
from circleseq import circleseq

TEST_OUTPUT_PATH = 'tmp'

TEST_MANIFEST_PATH = os.path.join('CIRCLEseq_StandardTest.yaml')

CORRECT_ALIGNED_OUTPUT = 'data/StandardOutput/aligned'
CORRECT_IDENTIFIED_OUTPUT = 'data/StandardOutput/identified'
CORRECT_VARIANTS_OUTPUT = 'data/StandardOutput/variants'
CORRECT_VISUALIZATION_OUTPUT = 'data/StandardOutput/visualization'

CORRECT_ALL_OUTPUT = 'data'

class FullPipelineTestCase(unittest.TestCase):

    def setUp(self):
        pass

    def testFullPipeline(self):
        c = circleseq.CircleSeq()
        c.parseManifest(TEST_MANIFEST_PATH)
        
        # Align and test the alignment output
        c.alignReads()
        self.assertTrue(utils.checkFolderEquality(os.path.join(c.analysis_folder, "aligned"), CORRECT_ALIGNED_OUTPUT))

        # Find cleavage sites
        c.findCleavageSites()
        self.assertTrue(utils.checkFolderEquality(os.path.join(c.analysis_folder, 'identified'), CORRECT_IDENTIFIED_OUTPUT))

        # Visualize filtered sites
        c.visualize()
        self.assertTrue(utils.checkFolderEquality(os.path.join(c.analysis_folder, 'visualization'), CORRECT_VISUALIZATION_OUTPUT))

        # Look for genomic variants
        c.callVariants()
        self.assertTrue(utils.checkFolderEquality(os.path.join(c.analysis_folder, 'variants'), CORRECT_VARIANTS_OUTPUT))


    def tearDown(self):
        pass

if __name__ == '__main__':
    unittest.main()


from __future__ import print_function
import argparse
import itertools
import gzip
from utility import reverseComplement, fq

def mergeReads(fastq1_filename, fastq2_filename, out):
    fastq1_file = fq(fastq1_filename)
    fastq2_file = fq(fastq2_filename)

    with gzip.open(out, 'wb') as o:
        for r1, r2 in itertools.izip(fastq1_file, fastq2_file):
            merged_sequence = reverseComplement(r1[1]) + r2[1]
            merged_quality_scores = r1[3][::-1] + r2[3]
            print(r1[0], file=o)
            print(merged_sequence, file=o)
            print(r1[2], file=o)
            print(merged_quality_scores, file=o)

def main():
    parser = argparse.ArgumentParser(description='Merge CIRCLE-seq reads for alignment.')
    parser.add_argument('--read1', help='Read 1 filename', required=True)
    parser.add_argument('--read2', help='Read 2 filename', required=True)
    parser.add_argument('--out', help='Output filename', required=True)

    args = parser.parse_args()

    mergeReads(args.read1, args.read2, args.out)

if __name__ == "__main__":
    main()


#!/usr/bin/env python
#-*- coding: utf-8 -*-

"""
circleseq.py as the wrapper for CIRCLE-seq analysis
"""

from alignReads import alignReads
from visualization import visualizeOfftargets
from mergeReads import mergeReads
import argparse
import os
import sys
import subprocess
import traceback
import log
import yaml
import validation
import findCleavageSites
import callVariants

logger = log.createCustomLogger('root')

class CircleSeq:

    def __init__(self):
        self.search_radius = 20
        self.window_size = 3
        self.mapq_threshold = 50
        self.start_threshold = 1
        self.gap_threshold = 3
        self.mismatch_threshold = 6
        self.read_threshold = 6
        self.merged_analysis = True
        self.all_chromosomes = False
        self.variant_analysis = False

    def parseManifest(self, manifest_path, sample='all'):
        logger.info('Loading manifest...')

        with open(manifest_path, 'r') as f:
            manifest_data = yaml.load(f)

        try:
            # Validate manifest data
            validation.validateManifest(manifest_data)

            self.BWA_path  = manifest_data['bwa']
            self.reference_genome = manifest_data['reference_genome']
            self.analysis_folder = manifest_data['analysis_folder']

            # Allow the user to specify read threshold, window_size and search_radius if they'd like
            if 'search_radius' in manifest_data:
                self.search_radius = manifest_data['search_radius']
            if 'window_size' in manifest_data:
                self.window_size = manifest_data['window_size']
            if 'mapq_threshold' in manifest_data:
                self.mapq_threshold = manifest_data['mapq_threshold']
            if 'start_threshold' in manifest_data:
                self.start_threshold = manifest_data['start_threshold']
            if 'gap_threshold' in manifest_data:
                self.gap_threshold = manifest_data['gap_threshold']
            if 'mismatch_threshold' in manifest_data:
                self.mismatch_threshold = manifest_data['mismatch_threshold']
            if 'read_threshold' in manifest_data:
                self.mismatch_threshold = manifest_data['read_threshold']
            if 'merged_analysis' in manifest_data:
                self.merged_analysis = manifest_data['merged_analysis']
            if 'all_chromosomes' in manifest_data:
                self.all_chromosomes = manifest_data['all_chromosomes']
            if 'variant_analysis' in manifest_data:
                self.variant_analysis = manifest_data['variant_analysis']
            # Allow the user to specify PAM seq. Yichao 4/29/2020
            if 'PAM' in manifest_data:
                self.PAM = manifest_data['PAM']
            else:
                self.PAM = "NGG"
            # Allow the user to specify Read Length. Yichao 4/29/2020
            if 'read_length' in manifest_data:
                self.read_length = manifest_data['read_length']
            else:
                self.read_length = 151
            # Allow the user to specify Read Count cutoff. Yichao 4/29/2020
            if 'read_count_cutoff' in manifest_data:
                self.read_count_cutoff = manifest_data['read_count_cutoff']
            else:
                self.read_count_cutoff = 6

            # Do not allow to run variant_analysis with merged_analysis
            if self.merged_analysis and self.variant_analysis:
                logger.error('merged_analysis is not compatible with variant_analysis. Please remove one option.')
                sys.exit()

            if sample == 'all':
                self.samples = manifest_data['samples']
            else:
                self.samples = {}
                self.samples[sample] = manifest_data['samples'][sample]
            # Make folders for output
            for folder in ['aligned', 'identified', 'fastq', 'visualization', 'variants']:
                output_folder = os.path.join(self.analysis_folder, folder)
                if not os.path.exists(output_folder):
                    os.makedirs(output_folder)

        except Exception as e:
            logger.error('Incorrect or malformed manifest file. Please ensure your manifest contains all required fields.')
            sys.exit()

    def alignReads(self):
        if self.merged_analysis:
            logger.info('Merging reads...')
            try:
                self.merged = {}
                for sample in self.samples:
                    sample_merge_path = os.path.join(self.analysis_folder, 'fastq', sample + '_merged.fastq.gz')
                    control_sample_merge_path = os.path.join(self.analysis_folder, 'fastq', 'control_' + sample + '_merged.fastq.gz')
                    mergeReads(self.samples[sample]['read1'],
                               self.samples[sample]['read2'],
                               sample_merge_path)
                    mergeReads(self.samples[sample]['controlread1'],
                               self.samples[sample]['controlread2'],
                               control_sample_merge_path)

                    sample_alignment_path = os.path.join(self.analysis_folder, 'aligned', sample + '.sam')
                    control_sample_alignment_path = os.path.join(self.analysis_folder, 'aligned', 'control_' + sample + '.sam')

                    alignReads(self.BWA_path,
                               self.reference_genome,
                               sample_merge_path,
                               '',
                               sample_alignment_path)

                    alignReads(self.BWA_path,
                               self.reference_genome,
                               control_sample_merge_path,
                               '',
                               control_sample_alignment_path)

                    self.merged[sample] = sample_alignment_path
                    logger.info('Finished merging and aligning reads.')

            except Exception as e:
                logger.error('Error aligning')
                logger.error(traceback.format_exc())
                quit()
        else:
            logger.info('Aligning reads...')
            try:
                self.aligned = {}
                self.aligned_sorted = {}
                for sample in self.samples:
                    sample_alignment_path = os.path.join(self.analysis_folder, 'aligned', sample + '.sam')
                    control_sample_alignment_path = os.path.join(self.analysis_folder, 'aligned', 'control_' + sample + '.sam')
                    alignReads(self.BWA_path,
                               self.reference_genome,
                               self.samples[sample]['read1'],
                               self.samples[sample]['read2'],
                               sample_alignment_path)
                    alignReads(self.BWA_path,
                               self.reference_genome,
                               self.samples[sample]['controlread1'],
                               self.samples[sample]['controlread2'],
                               control_sample_alignment_path)
                    self.aligned[sample] = sample_alignment_path
                    self.aligned_sorted[sample] = os.path.join(self.analysis_folder, 'aligned', sample + '_sorted.bam')
                    logger.info('Finished aligning reads to genome.')

            except Exception as e:
                logger.error('Error aligning')
                logger.error(traceback.format_exc())
                quit()

    def findCleavageSites(self):
        logger.info('Identifying off-target cleavage sites.')

        try:
            for sample in self.samples:
                if self.merged_analysis:
                    sorted_bam_file = os.path.join(self.analysis_folder, 'aligned', sample + '.bam')
                    control_sorted_bam_file = os.path.join(self.analysis_folder, 'aligned', 'control_' + sample + '.bam')
                else:
                    sorted_bam_file = os.path.join(self.analysis_folder, 'aligned', sample + '_sorted.bam')
                    control_sorted_bam_file = os.path.join(self.analysis_folder, 'aligned', 'control_' + sample + '_sorted.bam')
                identified_sites_file = os.path.join(self.analysis_folder, 'identified', sample)
                logger.info('Window: {0}, MAPQ: {1}, Gap: {2}, Start {3}, Mismatches {4}, Search_Radius {5}'.format(self.window_size, self.mapq_threshold, self.gap_threshold, self.start_threshold, self.mismatch_threshold, self.search_radius))
                findCleavageSites.compare(self.reference_genome, sorted_bam_file, control_sorted_bam_file, self.samples[sample]['target'],
                                          self.search_radius, self.window_size, self.mapq_threshold, self.gap_threshold,
                                          self.start_threshold, self.mismatch_threshold, sample, self.samples[sample]['description'],
                                          identified_sites_file, self.all_chromosomes, merged=self.merged_analysis)
        except Exception as e:
            logger.error('Error identifying off-target cleavage site.')
            logger.error(traceback.format_exc())
            quit()

    def visualize(self):
        logger.info('Visualizing off-target sites')

        # try:
            # for sample in self.samples:
                # if sample != 'control':
                    # infile = os.path.join(self.analysis_folder, 'identified', sample + '_identified_matched.txt')
                    # outfile = os.path.join(self.analysis_folder, 'visualization', sample + '_offtargets')
                    # visualizeOfftargets(infile, outfile, title=sample)

            # logger.info('Finished visualizing off-target sites')

        # except Exception as e:
            # logger.error('Error visualizing off-target sites.')
            # logger.error(traceback.format_exc())

        for sample in self.samples: ## 4/29/2020 Yichao solved: visualization stopped when sample has no off-target
            if sample != 'control':
                try:
                    infile = os.path.join(self.analysis_folder, 'identified', sample + '_identified_matched.txt')
                    outfile = os.path.join(self.analysis_folder, 'visualization', sample + '_offtargets')
                    visualizeOfftargets(infile, outfile, title=sample,PAM=self.PAM)
                except Exception as e:
                    logger.error('Error visualizing off-target sites: %s'%(sample))
                    logger.error(traceback.format_exc())
        logger.info('Finished visualizing off-target sites')


    def callVariants(self):

        try:
            if self.variant_analysis:
                logger.info('Identifying genomic variants')

                for sample in self.samples:
                    sorted_bam_file = os.path.join(self.analysis_folder, 'aligned', sample + '.bam')
                    identified_sites_file = os.path.join(self.analysis_folder, 'identified', sample + '_identified_matched.txt')
                    variants_basename = os.path.join(self.analysis_folder, 'variants', sample)
                    logger.info('Mismatches {0}, Search_Radius {1}'.format(self.mismatch_threshold, self.search_radius))
                    callVariants.getVariants(identified_sites_file, self.reference_genome, sorted_bam_file, variants_basename, self.search_radius, self.mismatch_threshold)

                logger.info('Finished identifying genomic variants')

        except Exception as e:
            logger.error('Error identifying genomic variants.')
            logger.error(traceback.format_exc())
            quit()

    def parallel(self, manifest_path, lsf, run='all'):
        logger.info('Submitting parallel jobs')
        current_script = __file__

        try:
            for sample in self.samples:
                cmd = 'python {0} {1} --manifest {2} --sample {3}'.format(current_script, run, manifest_path, sample)
                logger.info(cmd)
                subprocess.call(lsf.split() + [cmd])
            logger.info('Finished job submission')

        except Exception as e:
            logger.error('Error submitting jobs.')
            logger.error(traceback.format_exc())

    def referenceFree(self):
        pass

def parse_args():
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(description='Individual Step Commands',
                                       help='Use this to run individual steps of the pipeline',
                                       dest='command')

    all_parser = subparsers.add_parser('all', help='Run all steps of the pipeline')
    all_parser.add_argument('--manifest', '-m', help='Specify the manifest Path', required=True)
    all_parser.add_argument('--sample', '-s', help='Specify sample to process (default is all)', default='all')

    parallel_parser = subparsers.add_parser('parallel', help='Run all steps of the pipeline in parallel')
    parallel_parser.add_argument('--manifest', '-m', help='Specify the manifest Path', required=True)
    parallel_parser.add_argument('--lsf', '-l', help='Specify LSF CMD', default='bsub -R rusage[mem=32000] -P Genomics -q standard')
    parallel_parser.add_argument('--run', '-r', help='Specify which steps of pipepline to run (all, align, identify, visualize, variants)', default='all')

    align_parser = subparsers.add_parser('align', help='Run alignment only')
    align_parser.add_argument('--manifest', '-m', help='Specify the manifest Path', required=True)
    align_parser.add_argument('--sample', '-s', help='Specify sample to process (default is all)', default='all')

    merge_parser = subparsers.add_parser('merge', help='Merge paired end reads')
    merge_parser.add_argument('--manifest', '-m', help='Specify the manifest Path', required=True)
    merge_parser.add_argument('--sample', '-s', help='Specify sample to process (default is all)', default='all')

    identify_parser = subparsers.add_parser('identify', help='Run identification only')
    identify_parser.add_argument('--manifest', '-m', help='Specify the manifest Path', required=True)
    identify_parser.add_argument('--sample', '-s', help='Specify sample to process (default is all)', default='all')

    visualize_parser = subparsers.add_parser('visualize', help='Run visualization only')
    visualize_parser.add_argument('--manifest', '-m', help='Specify the manifest Path', required=True)
    visualize_parser.add_argument('--sample', '-s', help='Specify sample to process (default is all)', default='all')

    variants_parser = subparsers.add_parser('variants', help='Run variants analysis only')
    variants_parser.add_argument('--manifest', '-m', help='Specify the manifest Path', required=True)
    variants_parser.add_argument('--sample', '-s', help='Specify sample to process (default is all)', default='all')

    reference_free_parser = subparsers.add_parser('reference-free', help='Run reference-free discovery only')
    reference_free_parser.add_argument('--manifest', '-m', help='Specify the manifest Path', required=True)
    reference_free_parser.add_argument('--sample', '-s', help='Specify sample to process (default is all)', default='all')

    return parser.parse_args()

def main():
    args = parse_args()

    if args.command == 'all':
        c = CircleSeq()
        c.parseManifest(args.manifest, args.sample)
        c.alignReads()
        c.findCleavageSites()
        c.visualize()
        c.callVariants()
    elif args.command == 'parallel':
        c = CircleSeq()
        c.parseManifest(args.manifest)
        c.parallel(args.manifest, args.lsf, args.run)
    elif args.command == 'align':
        c = CircleSeq()
        c.parseManifest(args.manifest, args.sample)
        c.alignReads()
    elif args.command == 'identify':
        c = CircleSeq()
        c.parseManifest(args.manifest, args.sample)
        c.findCleavageSites()
    elif args.command == 'merge':
        c = CircleSeq()
        c.parseManifest(args.manifest, args.sample)
        c.mergeAlignReads()
    elif args.command == 'visualize':
        c = CircleSeq()
        c.parseManifest(args.manifest, args.sample)
        c.visualize()
    elif args.command == 'variants':
        c = CircleSeq()
        c.parseManifest(args.manifest, args.sample)
        c.callVariants()

if __name__ == '__main__':
    main()



"""
validation.py
=============

Contains utils for validating the filetype and existence of manifest-defined files/folders

"""

import logging
import os
import sys
from distutils.spawn import find_executable

logger = logging.getLogger('root')


def exists(filepath):
    if not os.path.isfile(filepath):
        logger.error('{0} does not exist'.format(filepath))
        sys.exit()


def checkIfBinary(filepath):
    executable = find_executable(filepath)

    if executable is None:
        logger.error('Executable binary not found at {0}'.format(filepath))
        sys.exit()

    # First check if file exists
    exists(executable)

    # Check if file is a valid binary
    # Adapted from http://stackoverflow.com/questions/898669/how-can-i-detect-if-a-file-is-binary-non-text-in-python
    textchars = bytearray({7,8,9,10,12,13,27} | set(range(0x20, 0x100)) - {0x7f})
    is_binary_string = lambda bytes: bool(bytes.translate(None, textchars))

    if not is_binary_string(open(executable, 'rb').read(1024)):
        logger.error('{0} is not a valid binary'.format(executable))
        sys.exit()


def checkIfFasta(filepath):
    # First check if file exists
    exists(os.path.abspath(filepath))


def checkIfFolder(folderpath):
    # Check if the folder exists
    if not os.path.isdir(os.path.abspath(folderpath)):
        logger.error('{0} is not a valid folder path'.format(folderpath))
        sys.exit()


def checkIfValidUndemultiplexed(undemultiplexed):
    # Check if read1, read2, index1, and index2 exist
    fields = ['forward', 'reverse', 'index1', 'index2']

    if set(fields) != set(undemultiplexed.keys()):
        logger.error('Undemultiplexed field must contain references to "forward", "reverse", "index1", "index2"')
        sys.exit()

    invalid_file = False
    for field in fields:
        if not os.path.isfile(undemultiplexed[field]):
            logger.error('"read1" undemultiplexed field does not reference a valid file')
            invalid_file = True

    if invalid_file:
        sys.exit()


def checkIfValidSamples(samples):
    # # Check if control is one of the samples
    # if 'control' not in samples:
    #     logger.error('A control sample must be specified')
    #     sys.exit()

    if len(samples.keys()) == 0:
        logger.error('No samples defined')
        sys.exit()

    for sample in samples:
        if 'read1' not in samples[sample] or 'read2' not in samples[sample]:
            logger.error('read1 and read2 must be specified for {0} sample'.format(sample))
            sys.exit()
        if 'controlread1' not in samples[sample] or 'controlread2' not in samples[sample]:
            logger.error('controlread1 and controlread2 must be specified for {0} sample'.format(sample))
            sys.exit()
        if 'target' not in samples[sample]:
            logger.error('target sequence must be specified for {0} sample'.format(sample))
            sys.exit()

def validateManifest(manifest_data):
    # Check if manifest contains the required fields
    fields = ['bwa', 'reference_genome', 'analysis_folder', 'samples']
    missing_fields = False

    for field in fields:
        if field not in manifest_data.keys():
            logger.error('"{0}" field must be specified in manifest'.format(field))
            missing_fields = True

    if missing_fields:
        sys.exit()

    # Now validate each field
    checkIfBinary(manifest_data['bwa'])
    checkIfBinary(manifest_data['samtools'])
    checkIfFasta(manifest_data['reference_genome'])
    checkIfValidSamples(manifest_data['samples'])


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# from distutils.core import setup
from setuptools import setup, find_packages

## conda skeleton can't find readme
import os
if os.path.isfile("README.MD"):
	with open("README.MD", "r") as fh:
		long_description = fh.read()
else:
	long_description="change-seq"

setup(
	name='changeseq',
	version='1.2.3',
	description="Bioinformatic pipeline for the CHANGE-seq assay.",
	author="Shengdar Q Tsai, Martin Aryee, Ved V Topkar, Jose Malagon-Lopez",
	author_email='STSAI4@mgh.harvard.edu, Aryee.Martin@mgh.harvard.edu, vedtopkar@gmail.com, jose.lopez@mail.harvard.edu',
	url='https://github.com/tsailabSJ/changeseq',
	packages=['changeseq'],
	# package_dir={'changeseq':'changeseq'},
	license='LICENSE',
	scripts=['changeseq/changeseq.py','changeseq/alignReads.py','changeseq/visualization.py',
		'changeseq/callVariants.py','changeseq/findCleavageSites.py','changeseq/log.py',
		'changeseq/mergeReads.py','changeseq/referenceFree.py','changeseq/utility.py',
		'changeseq/validation.py'],
	package_data={'test': ["test/*"]},
	include_package_data=True,
	long_description=long_description,
	long_description_content_type='text/markdown'	,
	keywords='changeseq',
	classifiers=[
		'Development Status :: 4 - Beta',
		'Intended Audience :: Science/Research',
		'Topic :: Scientific/Engineering :: Bio-Informatics',
		'Topic :: Scientific/Engineering :: Visualization',
		'Topic :: Scientific/Engineering :: Information Analysis',
		'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
		'Operating System :: Unix',
		'Natural Language :: English',
		"Programming Language :: Python :: 2",
		'Programming Language :: Python :: 2.6',
		'Programming Language :: Python :: 2.7'
	]
)



from __future__ import print_function

import argparse
import itertools
import regex
import re
import gzip
import sys
import collections
from findCleavageSites import regexFromSequence, alignSequences, reverseComplement, extendedPattern, realignedSequences

"""
FASTQ generator function from umi package
"""
def fq(file):
    if re.search('.gz$', file):
        fastq = gzip.open(file, 'rb')
    else:
        fastq = open(file, 'r')
    with fastq as f:
        while True:
            l1 = f.readline()
            if not l1:
                break
            l2 = f.readline()
            l3 = f.readline()
            l4 = f.readline()
            yield [l1, l2, l3, l4]

"""
Main function to find off-target sites in reference-free fashion
"""
def analyze(fastq1_filename, fastq2_filename, targetsite, out_base, name='', cells='', mismatch_threshold=7):

    read_count = 0
    c = collections.Counter()
    d = collections.defaultdict(list)

    fastq1_file = fq(fastq1_filename)
    fastq2_file = fq(fastq2_filename)
    for r1, r2 in itertools.izip(fastq1_file, fastq2_file):
        r1_sequence = r1[1].rstrip('\n')
        r2_sequence = r2[1].rstrip('\n')
        joined_seq = reverseComplement(r1_sequence) + r2_sequence
        truncated_joined_seq = joined_seq[130:170]

        sequence_data = alignSequences(targetsite, truncated_joined_seq, max_score=mismatch_threshold)
        offtarget, mismatch, length, strand, start, end, realigned_target = sequence_data[:7]

        if offtarget:
            c[offtarget] += 1
            d[offtarget].append(joined_seq)

        read_count += 1
        if not read_count % 100000:
            print(read_count/float(1000000), end=" ", file=sys.stderr)

    print('Finished tabulating reference-free discovery counts.', file=sys.stderr)
    out_filename = out_base + '.txt'

    with open(out_filename, 'w') as o:
        for target_sequence, target_count in c.most_common():
            print(target_sequence, target_count, file=o)
            off_target_fasta_filename = '{0}_{1:04d}_{2}.fasta'.format(out_base, target_count, target_sequence)
            with open(off_target_fasta_filename, 'w') as off_target_fasta_file:
                j = 0
                for sequence in d[target_sequence]:
                    j += 1
                    print('>{0:04d}_{1}_{2}'.format(target_count, target_sequence, j), file=off_target_fasta_file)
                    print(sequence, file=off_target_fasta_file)

def join_write_output(fastq1_filename, fastq2_filename, out):
    fastq1_file = fq(fastq1_filename)
    fastq2_file = fq(fastq2_filename)

    with open(out, 'w') as o:
        for r1, r2 in itertools.izip(fastq1_file, fastq2_file):
            header = '>{0}'.format(r1[0])
            r1_sequence = r1[1].rstrip('\n')
            r2_sequence = r2[1].rstrip('\n')
            joined_seq = reverseComplement(r1_sequence) + r2_sequence
            print(header, end='', file=o)
            print(joined_seq, file=o)


def main():
    parser = argparse.ArgumentParser(description='Identify off-target candidates from Illumina short read sequencing data.')
    parser.add_argument('--fq1', help='FASTQ Read 1', required=True)
    parser.add_argument('--fq2', help='FASTQ Read 2', required=True)
    parser.add_argument('--targetsite', help='Targetsite Sequence', required=True)
    parser.add_argument('--name', help='Targetsite Name', required=False)
    parser.add_argument('--cells', help='Cells', required=False)
    parser.add_argument('--mismatch_threshold', help='Maximum score threshold', default=7, type=int)
    parser.add_argument('--out', help='Output file base', required=True)
    args = parser.parse_args()

    analyze(args.fq1, args.fq2, args.targetsite, args.out, args.name, args.cells, args.mismatch_threshold)

if __name__ == "__main__":
    main()



from __future__ import print_function

import subprocess
import sys
import os
import argparse
import regex
import re
import HTSeq
import pyfaidx
from findCleavageSites import get_sequence, regexFromSequence, alignSequences, reverseComplement, extendedPattern, realignedSequences


"""
Run samtools:mpileup and get all identified variants in the window sequences
"""
def snpCall(matched_file, reference, bam_file, out, search_radius):
    basename = os.path.basename(out)
    output_folder = os.path.dirname(out)

    # open matched file
    regions = list()
    with open(matched_file, 'rU') as f:
        f.readline()
        for line in f:
            site = line.strip().split('\t')
            #  chromosome, windowStart, windowEnd, strand, bam, region_basename (=Targetsite_Name)
            regions.append([site[0], int(site[6]) - search_radius, int(site[7]) + search_radius, '*', bam_file, '_'.join([site[26], site[3]])])

    print('Running samtools:mpileup for %s' % basename, file=sys.stderr)
    out_vcf = os.path.join(output_folder, basename + '_mpileup_output')
    if os.path.exists(out_vcf):
        subprocess.check_call('rm -r %s' % out_vcf, shell=True, env=os.environ.copy())
    os.makedirs(out_vcf)
    process_mpileup = open(os.path.join(out_vcf, 'logFile_mpileup'), 'w')

    for item in regions:
        chromosome, windowStart, windowEnd, strand, bam_file, region_basename = item
        region = '%s%s%s%s%s' % (chromosome, ":", int(windowStart), "-", int(windowEnd))
        output = os.path.join(out_vcf, region_basename + '.vcf')

        cl_vcf = 'samtools mpileup -v --region %s --fasta-ref %s %s > %s' % (region, reference, bam_file, output)
        subprocess.check_call(cl_vcf, shell=True, env=os.environ.copy(), stderr=process_mpileup, stdout=process_mpileup)
    process_mpileup.close()

    print('Collecting variants for %s' % basename, file=sys.stderr)
    out_bcf = os.path.join(output_folder, basename + '_output_bcftools')
    if os.path.exists(out_bcf):
        subprocess.check_call('rm -r %s' % out_bcf, shell=True, env=os.environ.copy())
    os.makedirs(out_bcf)
    process_bcftools = open(os.path.join(out_bcf, 'logFile_bcftools'), 'w')

    vcf_files = [f for f in os.listdir(out_vcf) if os.path.isfile(os.path.join(out_vcf, f))]
    for arch in vcf_files:
        if not arch.startswith('.') and arch.endswith('.vcf'):
            name = arch[:-4]
            output = os.path.join(out_bcf, name + '_BCFcall.vcf')

            cl_bcf = 'bcftools call -v -c %s > %s' % (os.path.join(out_vcf, arch), output)
            subprocess.check_call(cl_bcf, shell=True, env=os.environ.copy(), stderr=process_bcftools, stdout=process_bcftools)
    process_bcftools.close()

    print('Collecting significant variant calls for %s' % basename, file=sys.stderr)
    out_svc = os.path.join(output_folder, basename + '_output_svc')
    if os.path.exists(out_svc):
        subprocess.check_call('rm -r %s' % out_svc, shell=True, env=os.environ.copy())
    os.makedirs(out_svc)
    process_svc = open(os.path.join(out_svc, 'logFile_svc'), 'w')

    bcf_files = [f for f in os.listdir(out_bcf) if os.path.isfile(os.path.join(out_bcf, f))]
    for arch in bcf_files:
        if not arch.startswith('.') and arch.endswith('.vcf'):
            name = arch[:-12]
            output = os.path.join(out_svc, name + '_SIGNFcall.txt')

            cl_sed = "sed -n '/##/!p' %s | awk 'FNR>1' > %s" % (os.path.join(out_bcf, arch), output)
            subprocess.check_call(cl_sed, shell=True, env=os.environ.copy(), stderr=process_svc, stdout=process_svc)
    process_svc.close()

    print('Consolidating all the significant variant calls for %s' % basename, file=sys.stderr)
    header = ['targetsite', 'site_name', 'chromosome', 'one_based_position', 'reference', 'variant', 'quality', 'genotype', 'depth', 'PL']
    variants = list()

    svc_files = [f for f in os.listdir(out_svc) if os.path.isfile(os.path.join(out_svc, f))]
    for arch in svc_files:
        if not arch.startswith('.') and arch.endswith('.txt'):
            tag = arch[:-14]
            f = open(os.path.join(out_svc, arch), 'r')
            reads = f.readlines()
            f.close()

            for line in reads:
                item = line.split()
                if 'INDEL' in item[7]:
                    variants.append(
                        [basename, tag] + item[:2] + item[3:6] + [str(int(item[9][0])) + '|' + str(int(item[9][2]))] +
                        [item[7].split(';')[3][3:]] + ['_'.join(item[9][4:].split(','))])
                else:
                    variants.append(
                        [basename, tag] + item[:2] + item[3:6] + [str(int(item[9][0])) + '|' + str(int(item[9][2]))] +
                        [item[7].split(';')[0][3:]] + ['_'.join(item[9][4:].split(','))])

    out_file = open(out + '_mpileupCall.txt', 'w')
    print(*header, sep='\t', file=out_file)
    for item in variants:
        print(*item, sep='\t', file=out_file)
    out_file.close()

    print('Cleaning up directive for %s' % basename, file=sys.stderr)
    subprocess.check_call('rm -r %s' % out_vcf, shell=True, env=os.environ.copy())
    subprocess.check_call('rm -r %s' % out_bcf, shell=True, env=os.environ.copy())
    subprocess.check_call('rm -r %s' % out_svc, shell=True, env=os.environ.copy())

    print('Done running samtools:mpileup for %s' % basename, file=sys.stderr)
    return variants


"""
Obtain variant off-target sequences
"""
def realignVariantBulge(bulge_sequence, window_sequence_variant, bulge_strand):
    bseq = bulge_sequence.replace('-', '')
    if bulge_strand == '+':
        m_bulge = re.search(bseq, window_sequence_variant, re.I)
    else:
        m_bulge = re.search(bseq, reverseComplement(window_sequence_variant), re.I)
    variant_bseq = m_bulge.group()
    variant_bseq = variant_bseq[:bulge_sequence.find('-')] + '-' + variant_bseq[bulge_sequence.find('-'):]
    return variant_bseq


def SNPreader(snp_file):
    ga = HTSeq.GenomicArray("auto", stranded=False, typecode='O')

    for snp in snp_file:
        basename, snpID, chromosome, one_based_position, reference, variant, quality, genotype, depth, PL = snp
        position = int(one_based_position) - 1
        key = '_'.join([basename, chromosome])
        ga[HTSeq.GenomicInterval(chromosome, position, position + 1, ".")] = [position, reference, variant, genotype, quality, key]
    return ga


def arrayOffTargets(matched_file, search_radius):
    offtargets_dict = {}
    gi_dict = {}

    with open(matched_file, 'r') as g:
        g.readline()
        for line in g:
            site = line.strip().split('\t')

            Chromosome = site[0]
            start = int(site[6]) - search_radius
            end = int(site[7]) + search_radius
            Name = site[3]

            offtargets_dict[Name] = site

            #  create a genomic interval for each window sequence
            gi_dict[Name] = HTSeq.GenomicInterval(Chromosome, start, end, ".")
    return offtargets_dict, gi_dict


def snpAdjustment(matched_file, snp_file, out, mismatch_threshold, search_radius):
    output_file = open(out + '_Variants.txt', 'w')
    print('Chromosome', 'Start', 'End', 'Name', 'ReadCount', 'Strand',
          'Variant_WindowSequence',
          'Variant_Site_SubstitutionsOnly.Sequence', 'Variant_Site_SubstitutionsOnly.NumSubstitutions',
          'Variant_Site_SubstitutionsOnly.Strand',
          'Variant_Site_GapsAllowed.Sequence', 'Variant_Site_GapsAllowed.Length', 
          'Variant_Site_GapsAllowed.Substitutions', 'Variant_Site_GapsAllowed.Insertions', 'Variant_Site_GapsAllowed.Deletions',
          'Variant_Site_GapsAllowed.Strand',
          'Cell', 'Targetsite', 'TargetSequence', 'Variant_RealignedTargetSequence',
          'Reference', 'Variant', 'Genotype', 'Quality',
          sep='\t', file=output_file)
    output_file.close()

    basename = os.path.basename(out)
    offtargets, gi_offtargets = arrayOffTargets(matched_file, search_radius)
    ga_snp = SNPreader(snp_file)

    for name in offtargets:
        variant_flag = False
        site = offtargets[name]
        gi = gi_offtargets[name]

        chromosome = site[0]
        window_sequence = site[9]
        window_sequence = window_sequence.upper()
        cell, targetsite = site[25:27]
        TargetSequence = site[28]
        output01 = site[0:6]
        output03 = [cell, targetsite, TargetSequence]
        ots_nb, ots_bu = site[10], site[15]

        #  obtain variant window sequence
        wkey = '_'.join([basename, chromosome])
        insert_start, insert_end, insert_var, snp_data = list(), list(), list(), {}

        for i, v in ga_snp[gi].steps():
            if v:
                position, reference, variant, genotype, quality, key = v
                if key == wkey:
                    variant = variant.split(',')[0]
                    for n, pos in enumerate(range(gi.start, gi.end)):
                        if pos == int(position):
                            insert_var.append(variant.lower())
                            insert_start.append(n)
                            end_pos = n + len(reference)
                            insert_end.append(end_pos)
                            snp_data[str(position)] = [position, reference, variant, genotype, quality]

        tri = 0
        window_sequence_variant = ''
        for i in range(len(insert_var)):
            variant = insert_var[i]
            pos = insert_start[i]
            window_sequence_variant += window_sequence[tri:pos] + variant.lower()
            tri = insert_end[i]
        window_sequence_variant += window_sequence[tri:]

        #  variant off-target sequences: only proceed if there is a variant in the window sequence
        window_sequence_var = window_sequence_variant.upper()
        if window_sequence_var != window_sequence:
            offtarget_sequence_no_bulge, mismatches, offtarget_sequence_length, chosen_alignment_strand_m, start_no_bulge, end_no_bulge, \
            realigned_target, \
            bulged_offtarget_sequence, length, score, substitutions, insertions, deletions, chosen_alignment_strand_b, bulged_start, bulged_end = \
                alignSequences(TargetSequence, window_sequence_var, max_score=mismatch_threshold)

            variant_ots_no_bulge, variant_ots_bulge = '', ''

            #  get variant sequence if the off-target sequences have changed by considering the variant window
            if ots_nb != offtarget_sequence_no_bulge:
                variant_flag = True
                if chosen_alignment_strand_m == '+':
                    m_no_bulge = re.search(offtarget_sequence_no_bulge, window_sequence_variant, re.I)
                else:
                    m_no_bulge = re.search(offtarget_sequence_no_bulge, reverseComplement(window_sequence_variant), re.I)
                variant_ots_no_bulge = m_no_bulge.group()

            if ots_bu != bulged_offtarget_sequence:
                variant_flag = True
                variant_ots_bulge = realignVariantBulge(bulged_offtarget_sequence, window_sequence_variant, chosen_alignment_strand_b)

            # collect and write variant data if we have variant off-target sequence(s)
            if variant_flag:
                total_genotype, total_reference, total_variant, total_quality = '', '', '', ''
                for pos in snp_data:
                    position, reference, variant, genotype, quality = snp_data[pos]
                    if total_genotype != '':
                        total_genotype += ''.join([':', genotype])
                        total_reference += ''.join([':', reference])
                        total_variant += ''.join([':', variant])
                        total_quality += ''.join([':', quality])
                    else:
                        total_genotype += ''.join([genotype])
                        total_reference += ''.join([reference])
                        total_variant += ''.join([variant])
                        total_quality += ''.join([quality])

                output02 = [variant_ots_no_bulge, mismatches, chosen_alignment_strand_m,
                            variant_ots_bulge, length, substitutions, insertions, deletions, chosen_alignment_strand_b]
                output04 = [total_reference, total_variant, total_genotype, total_quality]
                output_line = output01 + [window_sequence_variant] + output02 + output03 + [realigned_target] + output04

                with open(out + '_Variants.txt', 'a') as output_file:
                    print(*output_line, sep='\t', file=output_file)


"""
Main function
"""
def getVariants(matched_file, ref, bam_file, out, search_radius, mismatch_threshold):
    basename = os.path.basename(out)
    output_folder = os.path.dirname(out)
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    snp_file = snpCall(matched_file, ref, bam_file, out, search_radius)

    print('Obtaining Variant Off-Target Sequences for %s' % basename, file=sys.stderr)
    snpAdjustment(matched_file, snp_file, out, mismatch_threshold, search_radius)


def main():
    parser = argparse.ArgumentParser(description='Implement samtools:mpileup to identify genomic variants and adjust the off-target sequence when required.')
    parser.add_argument('--matched_file', help="full_path_to/matched file in 'identified' folder", required=True)
    parser.add_argument('--ref', help="Reference Genome Fasta", required=True)
    parser.add_argument('--bam', help="Sorted BAM file", required=True)
    parser.add_argument('--search_radius', help="Search radius around the position window", default=20, type=int)
    parser.add_argument('--mismatch_threshold', help='Maximum score threshold', default=7, type=int)
    parser.add_argument('--out', help="Output file basename, with full path", required=True)
    args = parser.parse_args()

    getVariants(args.matched_file, args.ref, args.bam, args.out, args.search_radius, args.mismatch_threshold)

if __name__ == "__main__":
    main()
